package org.processmining.plugins.cosa;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import java.util.Map;

import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.elements.ExpandableSubNet;
import org.processmining.models.graphbased.directed.petrinet.elements.Place;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;

/**
 * @author hverbeek
 * 
 *         ACTIVITY ::= token COORDINATE MEMO* ["input" CONNECTION*] ["output"
 *         CONNECTION*] ["sub_flow" SUBFLOW] ["process" PROCESS*] ["next_owner"
 *         token] ["access" ACCESS*] ("target" TARGET)* ["priority" token]
 *         ["with" WITH] "end" ["activity"|"call_activity"] ["declare"]
 */
public class CosaActivity extends CosaElement {

	private static boolean useSubnets = false;

	private String name;
	private CosaCoordinate coordinate;
	private final List<CosaMemo> memoList;
	private final List<CosaConnection> inputList;
	private final List<CosaConnection> outputList;
	private CosaSubflow subflow;
	private final List<CosaProcess> processList;
	//	private String nextOwner;
	private final List<CosaAccess> accessList;
	private final List<CosaTarget> targetList;
	//	private String priority;
	private CosaWith with;

	public CosaActivity() {
		name = null;
		coordinate = null;
		memoList = new ArrayList<CosaMemo>();
		inputList = new ArrayList<CosaConnection>();
		outputList = new ArrayList<CosaConnection>();
		subflow = null;
		processList = new ArrayList<CosaProcess>();
		//		nextOwner = null;
		accessList = new ArrayList<CosaAccess>();
		targetList = new ArrayList<CosaTarget>();
		//		priority = null;
		with = null;
	}

	public void parse(CosaStream input, CosaScript parent, CosaScript script, String path)
			throws FileNotFoundException, IOException {
		if (input.getToken().equals("activity") || input.getToken().equals("call_activity")) {
			input.advance();
		}
		name = input.getToken();
		input.advance();
		Collection<String> keywords = new ArrayList<String>(Arrays.asList("memo", "input", "output", "sub_flow",
				"process", "next_owner", "access", "target", "priority", "with", "end"));
		if (!keywords.contains(input.getToken())) {
			coordinate = new CosaCoordinate();
			coordinate.parse(input);
		}
		while (input.getToken().equals("memo")) {
			input.advance();
			CosaMemo memo = new CosaMemo();
			memo.parse(input, keywords);
			memoList.add(memo);
		}
		keywords.remove("memo");
		keywords.remove("input");
		if (input.getToken().equals("input")) {
			input.advance();
			while (!keywords.contains(input.getToken())) {
				CosaConnection connection = new CosaConnection();
				connection.parse(input);
				inputList.add(connection);
			}
		}
		keywords.remove("output");
		if (input.getToken().equals("output")) {
			input.advance();
			while (!keywords.contains(input.getToken())) {
				CosaConnection connection = new CosaConnection();
				connection.parse(input);
				outputList.add(connection);
			}
		}
		keywords.remove("sub_flow");
		if (input.getToken().equals("sub_flow")) {
			input.advance();
			subflow = new CosaSubflow();
			subflow.parse(input, script, path);
		}
		keywords.remove("process");
		if (input.getToken().equals("process")) {
			input.advance();
			while (!keywords.contains(input.getToken())) {
				CosaProcess process = new CosaProcess();
				process.parse(input);
				processList.add(process);
				if (input.getToken().equals("undo_process_begin")) {
					input.advance();
					while (!input.getToken().equals("endo_process_end")) {
						input.advance();
					}
					input.advance();
				}
			}
		}
		keywords.remove("next_owner");
		if (input.getToken().equals("next_owner")) {
			input.advance();
			//			nextOwner = input.getToken();
			input.advance();
		}
		keywords.remove("access");
		if (input.getToken().equals("access")) {
			input.advance();
			while (!keywords.contains(input.getToken())) {
				CosaAccess access = new CosaAccess();
				access.parse(input);
				accessList.add(access);
			}
		}
		while (input.getToken().equals("target")) {
			input.advance();
			CosaTarget target = new CosaTarget();
			target.parse(input, keywords);
			targetList.add(target);
		}
		keywords.remove("target");
		keywords.remove("priority");
		if (input.getToken().equals("priority")) {
			input.advance();
			//			priority = input.getToken();
			input.advance();
		}
		keywords.remove("with");
		if (input.getToken().equals("with")) {
			input.advance();
			with = new CosaWith();
			with.parse(input);
		}
		if (input.getToken().equals("end")) {
			input.advance();
		}
		if (input.getToken().equals("activity") || input.getToken().equals("call_activity")) {
			input.advance();
		}
		if (input.getToken().equals("declare")) {
			input.advance();
		}
	}

	public String getName() {
		return name;
	}

	public void convertToNet(PetrinetGraph net, ExpandableSubNet subnet, Place startPlace, Place endPlace,
			boolean isStartActivity, boolean isEndActivity, Map<String, Place> placeMap) {
		ExpandableSubNet activityNet = useSubnets ? net.addGroup(name, subnet) : subnet;
		/*
		 * Place
		 */
		Place startedPlace = net.addPlace(name + " started", activityNet);
		Place completedPlace = net.addPlace(name + " completed", activityNet);
		/*
		 * Thing in the middle (possibly an entire subflow)
		 */
		if (subflow != null) {
			ExpandableSubNet subsubnet = useSubnets ? net.addGroup(name + " subflow", activityNet) : subnet;
			subflow.convertToNet(net, subsubnet, startedPlace, completedPlace);
		} else {
			Transition transition = net.addTransition(name, activityNet);
			net.addArc(startedPlace, transition, activityNet);
			net.addArc(transition, completedPlace, activityNet);
		}
		/*
		 * Join
		 */
		for (List<Place> inputSet : getInputSets(isStartActivity, startPlace, placeMap)) {
			Transition transition = net.addTransition(name + " in", activityNet);
			transition.setInvisible(true);
			for (Place place : inputSet) {
				net.addArc(place, transition, subnet);
			}
			net.addArc(transition, startedPlace, activityNet);
		}
		/*
		 * Split
		 */
		for (List<Place> outputSet : getOutputSets(isEndActivity, endPlace, placeMap)) {
			Transition transition = net.addTransition(name + " out", activityNet);
			transition.setInvisible(true);
			net.addArc(completedPlace, transition, activityNet);
			for (Place place : outputSet) {
				net.addArc(transition, place, subnet);
			}
		}
	}

	private List<List<Place>> getInputSets(boolean isStartActivity, Place startPlace, Map<String, Place> placeMap) {
		CosaBehavior inputBehavior = getJoinBehavior();
		List<List<Place>> inputSets = new ArrayList<List<Place>>();
		List<Place> inputSet = new ArrayList<Place>();
		if (isStartActivity) {
			inputSet.add(startPlace);
			inputSets.add(inputSet);
		}
		for (CosaConnection connection : inputList) {
			if ((inputBehavior == CosaBehavior.AND) && !connection.hasPredicate()) {
				if (inputSets.isEmpty()) {
					inputSets.add(new ArrayList<Place>());
				}
				if (!connection.hasPredicate()) {
					for (List<Place> existingInputSet : inputSets) {
						existingInputSet.add(placeMap.get(connection.getName()));
					}
				}
			}
			if ((inputBehavior == CosaBehavior.OR)
					|| ((inputBehavior == CosaBehavior.AND) && connection.hasPredicate())) {
				if (inputSets.isEmpty()) {
					List<Place> newInputSet = new ArrayList<Place>();
					newInputSet.add(placeMap.get(connection.getName()));
					inputSets.add(newInputSet);
				} else {
					List<List<Place>> newInputSets = new ArrayList<List<Place>>();
					for (List<Place> existingInputSet : inputSets) {
						List<Place> newInputSet = new ArrayList<Place>(existingInputSet);
						newInputSet.add(placeMap.get(connection.getName()));
						newInputSets.add(newInputSet);
					}
					inputSets.addAll(newInputSets);
				}
			}
			if ((inputBehavior == CosaBehavior.OR) || (inputBehavior == CosaBehavior.XOR)) {
				List<Place> newInputSet = new ArrayList<Place>();
				newInputSet.add(placeMap.get(connection.getName()));
				inputSets.add(newInputSet);
			}
		}
		return inputSets;
	}

	private List<List<Place>> getOutputSets(boolean isEndActivity, Place endPlace, Map<String, Place> placeMap) {
		CosaBehavior outputBehavior = getSplitBehavior();
		List<List<Place>> outputSets = new ArrayList<List<Place>>();
		List<Place> outputSet = new ArrayList<Place>();
		if (isEndActivity) {
			outputSet.add(endPlace);
			outputSets.add(outputSet);
		}
		for (CosaConnection connection : outputList) {
			if ((outputBehavior == CosaBehavior.AND) && !connection.hasPredicate()) {
				if (outputSets.isEmpty()) {
					outputSets.add(new ArrayList<Place>());
				}
				if (!connection.hasPredicate()) {
					for (List<Place> existingOutputSet : outputSets) {
						existingOutputSet.add(placeMap.get(connection.getName()));
					}
				}
			}
			if ((outputBehavior == CosaBehavior.OR)
					|| ((outputBehavior == CosaBehavior.AND) && connection.hasPredicate())) {
				if (outputSets.isEmpty()) {
					List<Place> newOutputSet = new ArrayList<Place>();
					newOutputSet.add(placeMap.get(connection.getName()));
					outputSets.add(newOutputSet);
				} else {
					List<List<Place>> newOutputSets = new ArrayList<List<Place>>();
					for (List<Place> existingOutputSet : outputSets) {
						List<Place> newOutputSet = new ArrayList<Place>(existingOutputSet);
						newOutputSet.add(placeMap.get(connection.getName()));
						newOutputSets.add(newOutputSet);
					}
					outputSets.addAll(newOutputSets);
				}
			}
			if ((outputBehavior == CosaBehavior.OR) || (outputBehavior == CosaBehavior.XOR)) {
				List<Place> newOutputSet = new ArrayList<Place>();
				newOutputSet.add(placeMap.get(connection.getName()));
				outputSets.add(newOutputSet);
			}
		}
		return outputSets;
	}

	private CosaBehavior getJoinBehavior() {
		CosaBehavior behavior;
		for (CosaProcess process : processList) {
			behavior = process.getJoinBehavior();
			if (behavior != CosaBehavior.NN) {
				return behavior;
			}
		}
		return CosaBehavior.AND;
	}

	private CosaBehavior getSplitBehavior() {
		CosaBehavior behavior;
		for (CosaProcess process : processList) {
			behavior = process.getSplitBehavior();
			if (behavior != CosaBehavior.NN) {
				return behavior;
			}
		}
		return CosaBehavior.AND;
	}
}
