package org.processmining.plugins.cosa.importing;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;

import javax.swing.filechooser.FileFilter;
import javax.swing.filechooser.FileNameExtensionFilter;

import org.processmining.contexts.uitopia.annotations.UIImportPlugin;
import org.processmining.framework.abstractplugins.AbstractImportPlugin;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.plugin.annotations.PluginVariant;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinet.impl.PetrinetFactory;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.cosa.CosaScript;
import org.processmining.plugins.cosa.CosaStream;
import org.processmining.plugins.petrinet.reduction.Murata;

/**
 * @author hverbeek
 * 
 *         COSA import plug-in. Parses and imports a COSA script file and
 *         converts this to a marked Petri net.
 */
@Plugin(name = "Import Petri net from COSA file", parameterLabels = { "Filename" }, returnLabels = { "Petri Net",
		"Marking" }, returnTypes = { Petrinet.class, Marking.class })
@UIImportPlugin(description = "COSA Petri net files", extensions = { "scr" })
public class CosaImportNet extends AbstractImportPlugin {

	/*
	 * The folder where the script file resides. Needed in case there are
	 * subflows.
	 */
	private String path;

	protected FileFilter getFileFilter() {
		return new FileNameExtensionFilter("COSA files", "scr");
	}

	/*
	 * Override the default implementation of this method to be able to get to
	 * the file's folder. (non-Javadoc)
	 * 
	 * @see
	 * org.processmining.plugins.abstractplugins.AbstractImportPlugin#importFile
	 * (org.processmining.framework.plugin.PluginContext, java.io.File)
	 */
	@PluginVariant(requiredParameterLabels = { 0 })
	public Object importFile(PluginContext context, File f) throws Exception {
		/*
		 * Store the file's folder.
		 */
		path = f.getParent();
		/*
		 * Revert to the default implementation.
		 */
		return super.importFile(context, f);
	}

	protected Object importFromStream(PluginContext context, InputStream input, String filename, long fileSizeInBytes)
			throws FileNotFoundException, IOException {
		/*
		 * Create a fresh script.
		 */
		CosaScript script = new CosaScript();
		/*
		 * Parse the stream contents into the script.
		 */
		script.parse(new CosaStream(input), null, path);
		/*
		 * Create a fresh Petri net and a fresh marking.
		 */
		Petrinet net = PetrinetFactory.newPetrinet(script.getName() + " (imported from " + filename + ")");
		Marking marking = new Marking();
		/*
		 * Fill in the details of the Petrinet and the marking according to the
		 * script.
		 */
		script.convertToNet(net, null, marking);
		/*
		 * Create a fresh result.
		 */
		Object[] objects = reduce(context, net, marking);
		/*
		 * Add a connection from the Petri net to the marking.
		 */
		context.addConnection(new InitialMarkingConnection(net, marking));
		/*
		 * Set the label of the Petri net.
		 */
		context.getFutureResult(0).setLabel(net.getLabel());
		/*
		 * set the label of the marking.
		 */
		context.getFutureResult(1).setLabel("Marking of " + net.getLabel());
		/*
		 * Return the results.
		 */
		return objects;
	}

	/*
	 * Redcues tha given net and marking using the Murata rules.
	 */
	private Object[] reduce(PluginContext context, Petrinet net, Marking marking) {
		/*
		 * Get a Murata reducer.
		 */
		Murata murata = new Murata();
		Object objects[];
		try {
			/*
			 * Run the Murata reducer on the net net and marking.
			 */
			objects = murata.run(context, net, marking);
		} catch (ConnectionCannotBeObtained e) {
			/*
			 * Unable to reduce, return original net and marking.
			 */
			objects = new Object[2];
			objects[0] = net;
			objects[1] = marking;
		}
		return objects;
	}
}
